def rarefaction_curve_legacy(TaXon_table_xlsx, repetitions, path_to_outdirs, template, theme, font_size, taxonomic_level_1):

    import random
    import PySimpleGUI as sg
    import pandas as pd
    import numpy as np
    import plotly.graph_objects as go
    from pathlib import Path
    import webbrowser

    color1 = theme[0]
    color2 = theme[1]
    opacity_value = theme[2]

    ## create a y axis title text
    taxon_title = taxonomic_level_1.lower()

    ## adjust taxonomic level if neccessary
    if taxonomic_level_1 in ["ASVs", "ESVs", "OTUs", "zOTUs"]:
        taxon_title = taxonomic_level_1
        taxonomic_level_1 = "ID"

    TaXon_table_file = Path(TaXon_table_xlsx)

    TaXon_table_xlsx = pd.ExcelFile(TaXon_table_xlsx)
    df = pd.read_excel(TaXon_table_xlsx, 'TaXon table', header=0)
    df = df.replace(np.nan,"nan")

    available_samples = df.columns.tolist()[10:]
    sample_dict_clean = {}

    # iterate through all available samples
    for sample in available_samples:
        # create a dict for the read numbers of the respective sample for each species
        sample_OTU_list = df[[sample, taxonomic_level_1]].values.tolist()
        # select only the present Species
        sample_species_list = list(set([OTU[1] for OTU in sample_OTU_list if (OTU[0] != 0 and OTU[1] != "nan")]))
        # store the species in a dictionary
        sample_dict_clean[sample] = sample_species_list

    # draw once for each sample
    number_of_draws = len(sample_dict_clean.keys())

    # dictionary to store the drawing results
    draw_dictionary = {}

    ############################################################################
    ## create the progress bar window
    layout = [[sg.Text('Progress bar')],
              [sg.ProgressBar(1000, orientation='h', size=(20, 20), key='progressbar')],
              [sg.Cancel()]]
    window_progress_bar = sg.Window('Progress bar', layout, keep_on_top=True)
    progress_bar = window_progress_bar['progressbar']
    progress_update = 0
    progress_increase = 1000 / repetitions
    ############################################################################

    for n_reps in range(0, repetitions):
        # store the original dictionary to start over again
        # a copy of the original dictionary is required, because the samples will be removed with each draw
        # thus for each replicate a new dictionary to draw from has to be created
        sample_dict_to_draw = dict(sample_dict_clean)

        species_list = []
        species_set = []

        for i in range(0, number_of_draws):
            # choose a random sample from the dictionary
            random_choice = random.choice(list(sample_dict_to_draw.keys()))
            # extract the OTU IDs from the chosen sample and add them to the already existing OTU IDs
            species_list = species_list + sample_dict_clean[random_choice]
            # create a unique set
            species_set = set(species_list)
            # number of OTUs
            n_species = len(species_set)
            # now add the unique OTU list to the output dictionary
            # if the key is not in the dict, create a new entry (= OTU ID plus number of OTUs)
            if i not in draw_dictionary.keys():
                draw_dictionary[i] = [n_species]
            # if the key already exists, calculate the sum of the already existing number of OTUs and the new number of OTUs
            else:
                # create a new list to store the current number of OTUs
                add_species_list = draw_dictionary[i]
                add_species_list.append(n_species)
                draw_dictionary[i] = add_species_list

            # remove the sample to draw only once
            sample_dict_to_draw.pop(random_choice)

        ############################################################################
        event, values = window_progress_bar.read(timeout=1)
        if event == 'Cancel'  or event is None:
            window_progress_bar.Close()
            raise RuntimeError
        # update bar with loop value +1 so that bar eventually reaches the maximum
        progress_update += progress_increase
        progress_bar.UpdateBar(progress_update)
        ############################################################################

    window_progress_bar.Close()

    # create a dict to store the average number of OTUs per draw
    rarefaction_dict_average, rarefaction_dict_stdef = {}, {}

    def average(lst):
        return sum(lst) / len(lst)

    # iterate through the draw_dictionary and calculate the average number of OTUs
    for key, value in draw_dictionary.items():
        average_species = average(draw_dictionary[key])
        stdef_species = np.std(draw_dictionary[key], dtype=np.float64)
        rarefaction_dict_average[key] = average_species
        rarefaction_dict_stdef[key] = stdef_species

    # draw the plot
    draws = [i+1 for i in rarefaction_dict_average.keys()]
    n_species = list(rarefaction_dict_average.values())
    error_bar = list(rarefaction_dict_stdef.values())
    y_axis_title = "# " + taxon_title
    fig = go.Figure(data=[go.Scatter(x=draws, y=n_species, error_y=dict(type='data', array=error_bar, thickness=0.5, width=3, visible=True))])
    fig.update_layout(title_text="repetitions = " + str(n_reps+1), yaxis_title=y_axis_title, xaxis_title="# samples")
    fig.update_traces(marker_color=color1, marker_line_color=color2, opacity=opacity_value)
    fig.update_layout(height=800, width=1200, template=template, showlegend=False, font_size=font_size, title_font_size=font_size)

    ## write files
    output_pdf = Path(str(path_to_outdirs) + "/" + "Rarefaction_curves" + "/" + TaXon_table_file.name + "_rarefaction_" + taxon_title + ".pdf")
    output_html = Path(str(path_to_outdirs) + "/" + "Rarefaction_curves" + "/" + TaXon_table_file.name + "_rarefaction_taxa" + taxon_title + ".html")
    fig.write_image(str(output_pdf))
    fig.write_html(str(output_html))

    ## ask to show file
    answer = sg.PopupYesNo('Show plot?', keep_on_top=True)
    if answer == "Yes":
        webbrowser.open('file://' + str(output_html))

    ## print closing text
    closing_text = "Rarefaction curves are found in: " + str(path_to_outdirs) + "/rarefaction_curves/"
    sg.Popup(closing_text, title="Finished", keep_on_top=True)

    ## write log
    from taxontabletools.create_log import ttt_log
    ttt_log("rarefaction curve all-in-one", "analysis", TaXon_table_file.name, output_pdf.name, "nan", path_to_outdirs)

def rarefaction_curve_taxa(TaXon_table_xlsx, repetitions, path_to_outdirs, template, font_size, taxonomic_level_1, taxonomic_level_2, color_discrete_sequence):

    import random
    import PySimpleGUI as sg
    import pandas as pd
    import numpy as np
    import plotly.graph_objects as go
    import plotly.express as px
    from pathlib import Path
    import webbrowser

    ## load the TaXon table
    TaXon_table_file = Path(TaXon_table_xlsx)
    TaXon_table_xlsx = pd.ExcelFile(TaXon_table_xlsx)
    df = pd.read_excel(TaXon_table_xlsx, 'TaXon table', header=0)
    df = df.replace(np.nan,"nan")

    ## create a y axis title text
    taxon_title = taxonomic_level_1.lower()

    ## adjust taxonomic level if neccessary
    if taxonomic_level_1 in ["ASVs", "ESVs", "OTUs", "zOTUs"]:
        taxon_title = taxonomic_level_1
        taxonomic_level_1 = "ID"

    ## collect available samples
    available_samples = df.columns.tolist()[10:]
    sample_dict_clean = {}

    ## collect available taxa
    available_taxa = [taxon for taxon in set(df[taxonomic_level_2].values.tolist()) if taxon != "nan"]

    ## create a color dict
    ## extend to the color dict
    color_discrete_sequence = color_discrete_sequence * len(available_taxa)
    color_dict = {}
    for i, taxon in enumerate(available_taxa):
        color_dict[taxon] = color_discrete_sequence[i]

    ## collect the increase for each taxon
    increase_dict = {}

    ## create an empty figure
    fig = go.Figure()

    for taxon in sorted(available_taxa):
        df_filtered = df.loc[df[taxonomic_level_2] == taxon]

        # iterate through all available samples
        for sample in available_samples:
            # create a dict for the read numbers of the respective sample for each species
            sample_OTU_list = df_filtered[[sample, taxonomic_level_1]].values.tolist()
            # select only the present Species
            sample_species_list = list(set([OTU[1] for OTU in sample_OTU_list if (OTU[0] != 0 and OTU[1] != "nan")]))
            # store the species in a dictionary
            sample_dict_clean[sample] = sample_species_list

        # draw once for each sample
        number_of_draws = len(sample_dict_clean.keys())

        # dictionary to store the drawing results
        draw_dictionary = {}

        for n_reps in range(0, repetitions):
            # store the original dictionary to start over again
            # a copy of the original dictionary is required, because the samples will be removed with each draw
            # thus for each replicate a new dictionary to draw from has to be created
            sample_dict_to_draw = dict(sample_dict_clean)

            species_list = []
            species_set = []

            for i in range(0, number_of_draws):
                # choose a random sample from the dictionary
                random_choice = random.choice(list(sample_dict_to_draw.keys()))
                # extract the OTU IDs from the chosen sample and add them to the already existing OTU IDs
                species_list = species_list + sample_dict_clean[random_choice]
                # create a unique set
                species_set = set(species_list)
                # number of OTUs
                n_species = len(species_set)
                # now add the unique OTU list to the output dictionary
                # if the key is not in the dict, create a new entry (= OTU ID plus number of OTUs)
                if i not in draw_dictionary.keys():
                    draw_dictionary[i] = [n_species]
                # if the key already exists, calculate the sum of the already existing number of OTUs and the new number of OTUs
                else:
                    # create a new list to store the current number of OTUs
                    add_species_list = draw_dictionary[i]
                    add_species_list.append(n_species)
                    draw_dictionary[i] = add_species_list

                # remove the sample to draw only once
                sample_dict_to_draw.pop(random_choice)

        # create a dict to store the average number of OTUs per draw
        rarefaction_dict_average, rarefaction_dict_stdef = {}, {}

        def average(lst):
            return sum(lst) / len(lst)

        # iterate through the draw_dictionary and calculate the average number of OTUs
        for key, value in draw_dictionary.items():
            average_species = average(draw_dictionary[key])
            stdef_species = np.std(draw_dictionary[key], dtype=np.float64)
            rarefaction_dict_average[key] = average_species
            rarefaction_dict_stdef[key] = stdef_species

        ## add to plot
        draws = [i+1 for i in rarefaction_dict_average.keys()]
        n_species = list(rarefaction_dict_average.values())
        increase_dict[taxon] = n_species
        error_bar = list(rarefaction_dict_stdef.values())
        fig.add_trace(go.Scatter(x=draws, y=n_species, name=taxon, marker_color=color_dict[taxon], error_y=dict(type='data', array=error_bar, thickness=0.5, width=3, visible=True)))

    ## update figure
    y_axis_title = "# " + taxon_title
    fig.update_layout(title_text="repetitions = " + str(n_reps+1), yaxis_title=y_axis_title, xaxis_title="# samples")
    fig.update_layout(height=700, width=1200, template="simple_white", showlegend=True, font_size=font_size, title_font_size=font_size)
    fig.update_xaxes(rangemode="tozero")
    fig.update_yaxes(rangemode="tozero")
    fig.update_layout(height=800, width=1200, template=template, showlegend=True, font_size=font_size, title_font_size=font_size)

    ## write files
    out_name = taxonomic_level_1.lower() + "_" + taxonomic_level_2.lower()
    output_pdf = Path(str(path_to_outdirs) + "/" + "Rarefaction_curves" + "/" + TaXon_table_file.name + "_rarefaction_" + out_name + ".pdf")
    output_html = Path(str(path_to_outdirs) + "/" + "Rarefaction_curves" + "/" + TaXon_table_file.name + "_rarefaction_" + out_name + ".html")
    fig.write_image(str(output_pdf))
    fig.write_html(str(output_html))

    ## ask to show file
    answer = sg.PopupYesNo('Show plot?', keep_on_top=True)
    if answer == "Yes":
        webbrowser.open('file://' + str(output_html))

    ## print closing text
    closing_text = "Rarefaction curves are found in: " + str(path_to_outdirs) + "/rarefaction_curves/"
    sg.Popup(closing_text, title="Finished", keep_on_top=True)

    ## write log
    from taxontabletools.create_log import ttt_log
    ttt_log("rarefaction curve per taxon", "analysis", TaXon_table_file.name, output_pdf.name, "nan", path_to_outdirs)


def rarefaction_curve_reads(TaXon_table_xlsx, repetitions, width, height, path_to_outdirs, template, theme, font_size):

    import pandas as pd
    import PySimpleGUI as sg
    import numpy as np
    from statistics import mean
    from pathlib import Path
    import plotly.graph_objects as go
    from plotly.subplots import make_subplots
    import math, webbrowser

    TaXon_table_file = Path(TaXon_table_xlsx)
    TaXon_table_df = pd.read_excel(TaXon_table_xlsx).fillna("")
    samples = TaXon_table_df.columns.tolist()[10:]
    scatter_size = 5

    color1 = theme[0]
    color2 = theme[1]
    opacity_value = theme[2]

    height = int(height)

    ## count rows and columns to create subplots
    n_rows = math.ceil(len(samples) / 4)
    n_columns = 5
    column_count = 1
    row_count = 1
    fig = make_subplots(rows=n_rows, cols=4, subplot_titles=samples, shared_yaxes=True)

    ## calculate maximum number of OTUs
    max_OTUs = []
    for sample in samples:
        max_OTUs.append(len([OTU for OTU in TaXon_table_df[sample] if OTU != 0]))
    y_limit = max(max_OTUs) + 20

    ############################################################################
    ## create the progress bar window
    layout = [[sg.Text('Progress bar')],
              [sg.ProgressBar(1000, orientation='h', size=(20, 20), key='progressbar')],
              [sg.Cancel()]]
    window_progress_bar = sg.Window('Progress bar', layout, keep_on_top=True)
    progress_bar = window_progress_bar['progressbar']
    progress_update = 0
    progress_increase = 1000 / len(samples)
    ############################################################################

    ############################################################################
    event, values = window_progress_bar.read(timeout=1)
    if event == 'Cancel'  or event is None:
        window_progress_bar.Close()
        raise RuntimeError
    # update bar with loop value +1 so that bar eventually reaches the maximum
    progress_update += 0
    progress_bar.UpdateBar(progress_update)
    ############################################################################

    for sample in samples:

        ## filter sample from data
        read_df = TaXon_table_df[[sample, "ID"]]
        ## drop empty OTUs
        read_df = read_df[read_df[sample] != 0]
        ## create read list to draw the subsamples from
        read_list = pd.Series(np.repeat(read_df['ID'].to_list(), read_df[sample].to_list()))

        output = []

        ## draw random sample
        for perc in np.arange(0.00, 1.05, 0.05):
            ## calculate sample size
            sub_sample_size = int(len(read_list) * perc)

            ## draw X subsamples of that size
            mean_species = mean([read_list.sample(n = sub_sample_size).nunique() for i in range(repetitions)])

            output.append(mean_species)

        output = pd.DataFrame({'percentage': np.arange(0.00, 1.05, 0.05), 'mean_OTUs': output})

        ## write plot
        fig.add_trace(go.Scatter(x=output["percentage"], y=output["mean_OTUs"], name=sample, mode='markers+lines', marker=dict(size=int(scatter_size))), row=row_count, col=column_count)
        fig.update_traces(marker_color=color1, marker_line_color=color2, opacity=opacity_value, row=row_count, col=column_count)
        fig.update_yaxes(range=[0, y_limit], row=row_count, col=column_count)

        ## add a y axis title to all left bound plots
        if column_count == 1:
            fig.update_yaxes(title_text="# OTUs", row=row_count, col=column_count)

        ## add x axis title to all plots in the last row
        # if row_count == n_rows:
        #     fig.update_xaxes(title_text="subsample (%)", row=row_count, col=column_count)

        column_count += 1
        if column_count == n_columns:
            column_count = 1
            row_count += 1
            height += 100

        ############################################################################
        event, values = window_progress_bar.read(timeout=1)
        if event == 'Cancel'  or event is None:
            window_progress_bar.Close()
            raise RuntimeError
        # update bar with loop value +1 so that bar eventually reaches the maximum
        progress_update += progress_increase
        progress_bar.UpdateBar(progress_update)
        ############################################################################

    window_progress_bar.Close()

    fig.update_layout(height=int(height), width=int(width), template=template, font_size=font_size, title_font_size=font_size, showlegend=False)
    fig.update_yaxes(rangemode="tozero")
    fig.update_xaxes(rangemode="tozero")

    ## write files
    output_pdf = Path(str(path_to_outdirs) + "/" + "Rarefaction_curves" + "/" + TaXon_table_file.name + "_rarefaction_reads.pdf")
    output_html = Path(str(path_to_outdirs) + "/" + "Rarefaction_curves" + "/" + TaXon_table_file.name + "_rarefaction_reads.html")
    fig.write_image(str(output_pdf))
    fig.write_html(str(output_html))

    ## ask to show file
    answer = sg.PopupYesNo('Show plot?', keep_on_top=True)
    if answer == "Yes":
        webbrowser.open('file://' + str(output_html))

    ## print closing text
    closing_text = "Rarefaction curves are found in: " + str(path_to_outdirs) + "/rarefaction_curves/"
    sg.Popup(closing_text, title="Finished", keep_on_top=True)

    ## write log
    from taxontabletools.create_log import ttt_log
    ttt_log("rarefaction curve reads", "analysis", TaXon_table_file.name, output_pdf.name, repetitions, path_to_outdirs)



def calculate_taxonomic_resolution(TaXon_table_xlsx, path_to_outdirs, x_tax_res, y_tax_res, figure_type, template, theme, font_size, clustering_unit):

    import glob
    import PySimpleGUI as sg
    import pandas as pd
    from pandas import DataFrame
    import numpy as np
    import plotly.graph_objects as go
    from pathlib import Path
    import webbrowser

    color1 = theme[0]
    color2 = theme[1]
    opacity_value = theme[2]

    TaXon_table_file =  Path(TaXon_table_xlsx)
    TaXon_table_xlsx =  Path(TaXon_table_xlsx)
    TaXon_table_df = pd.read_excel(TaXon_table_xlsx)
    TaXon_table_df = TaXon_table_df.replace(np.nan, 'nan', regex=True)

    taxonomic_levels = ["Phylum", "Class", "Order", "Family", "Genus", "Species"]
    statistics_list, statistics_set, statistics_dict, highest_level_dict = [], [], {}, {}

    title = "# " + clustering_unit

    for taxon_to_evaluate in taxonomic_levels:
        taxa_list = [x for x in TaXon_table_df[taxon_to_evaluate].values.tolist() if str(x) != 'nan']
        statistics = taxon_to_evaluate, len(taxa_list)
        statistics_set.append(len(set(taxa_list)))
        statistics_list.append(list(statistics))
        statistics_dict[taxon_to_evaluate] = len(taxa_list)

    highest_level_dict["Phylum"] = statistics_dict["Phylum"] - statistics_dict["Class"]
    highest_level_dict["Class"] = statistics_dict["Class"] - statistics_dict["Order"]
    highest_level_dict["Order"] = statistics_dict["Order"] - statistics_dict["Family"]
    highest_level_dict["Family"] = statistics_dict["Family"] - statistics_dict["Genus"]
    highest_level_dict["Genus"] = statistics_dict["Genus"] - statistics_dict["Species"]
    highest_level_dict["Species"] = statistics_dict["Species"]

    taxon_levels = list(highest_level_dict.keys())
    highest_level_OTUs = list(highest_level_dict.values())
    total_OTUs = list(statistics_dict.values())

    # create plot
    # option A: scatter plot
    if figure_type == "a":

        fig = go.Figure(data=[go.Bar(x=taxon_levels, y=highest_level_OTUs, name="Taxon", textposition="outside", text=highest_level_OTUs)])
        fig.update_traces(marker_color=color1, marker_line_color=color2,marker_line_width=1, opacity=opacity_value)
        fig.update_layout(title_text='Taxonomic resolution (highest taxonomic level)', yaxis_title=title)
        fig.update_layout(height=int(y_tax_res), width=int(x_tax_res), template=template, font_size=font_size, title_font_size=font_size)

        ## finish script
        output_pdf = Path(str(path_to_outdirs) + "/" + "Taxonomic_resolution_plots" + "/" + TaXon_table_file.stem + "_taxonomic_resolution_a.pdf")
        output_html = Path(str(path_to_outdirs) + "/" + "Taxonomic_resolution_plots" + "/" + TaXon_table_file.stem + "_taxonomic_resolution_a.html")
        fig.write_image(str(output_pdf))
        fig.write_html(str(output_html))

        ## ask to show file
        answer = sg.PopupYesNo('Show plot?', keep_on_top=True)
        if answer == "Yes":
            webbrowser.open('file://' + str(output_html))

        ## write log file
        from taxontabletools.create_log import ttt_log
        ttt_log("taxonomic resolution", "analysis", TaXon_table_file.name, output_pdf.name, "plot a", path_to_outdirs)

    # option B: bar plot
    else:

        fig = go.Figure(data=[go.Bar(x=taxon_levels, y=total_OTUs, name="Taxon", textposition="outside", text=total_OTUs)])
        fig.update_traces(marker_color=color1, marker_line_color=color2,marker_line_width=1, opacity=opacity_value)
        fig.update_layout(title_text='Taxonomic resolution (total number of OTUs)', yaxis_title=title)
        fig.update_layout(height=int(y_tax_res), width=int(x_tax_res), template=template, font_size=font_size, title_font_size=font_size)

        ## finish script
        output_pdf = Path(str(path_to_outdirs) + "/" + "Taxonomic_resolution_plots" + "/" + TaXon_table_file.stem + "_taxonomic_resolution_b.pdf")
        output_html = Path(str(path_to_outdirs) + "/" + "Taxonomic_resolution_plots" + "/" + TaXon_table_file.stem + "_taxonomic_resolution_b.html")
        fig.write_image(str(output_pdf))
        fig.write_html(str(output_html))

        ## show plot
        answer = sg.PopupYesNo('Show plot?', keep_on_top=True)
        if answer == "Yes":
            webbrowser.open('file://' + str(output_html))

        ## write log file
        from taxontabletools.create_log import ttt_log
        ttt_log("taxonomic resolution", "analysis", TaXon_table_file.name, output_pdf.name, "plot b", path_to_outdirs)

    closing_text = "\n" + "Taxonomic resolution plots are found in: " + str(path_to_outdirs) + "/taxonomic_resolution_plots/"
    sg.Popup(closing_text, title="Finished", keep_on_top=True)



def ttt_log(task, type, input_file, output_file, parameters, path_to_outdirs):

    import os
    import pandas as pd
    import numpy as np
    from datetime import datetime
    from pathlib import Path

    dateTimeObj = datetime.now()
    time = str(dateTimeObj)

    log_xlsx = Path(str(path_to_outdirs) + "/log.xlsx")

    if os.path.exists(Path(log_xlsx)):
        log_df = pd.read_excel(log_xlsx)
        log_df = pd.DataFrame(log_df.values.tolist() + [[task, input_file, output_file, parameters, time, type]],columns=log_df.columns.tolist())
    else:
        log_df = pd.DataFrame([[task, input_file, output_file, parameters, time, type]],columns=["task", "input", "output", "parameters", "time", "type"])

    log_df.to_excel(log_xlsx, index=False)

def ttt_log_network(path_to_outdirs):

    import os
    import pandas as pd
    import numpy as np
    import networkx as nx
    import matplotlib.pyplot as plt
    import PySimpleGUI as sg
    from pathlib import Path

    log_xlsx = Path(str(path_to_outdirs) + "/log.xlsx")

    if not os.path.exists(Path(log_xlsx)):
        sg.PopupError("No log file exists!", title="Error", keep_on_top=True)
        raise RuntimeError

    win2_layout = [
                [sg.Text('',size=(1,1))],
                [sg.Text('Please adjust the plot size to the size of the network.')],
                [sg.Text("Plot size (w,h):"), sg.Input("12", size=(3,1), key="x_log"), sg.Input("12", size=(3,1), key="y_log")],
                [sg.Text('',size=(1,1))],
                [sg.Button('Continue', key='Continue')],
                ]

    win2 = sg.Window('Projects', win2_layout)
    event, values = win2.read()
    x_log = int(values["x_log"])
    y_log = int(values["y_log"])
    win2.close()

    df = pd.read_excel(log_xlsx)

    edge_label_dict = {}
    node_label_dict = {}
    for task in df.values.tolist():
        # collect data from the log table
        input = task[1]
        output = task[2]
        edge_label = task[0]
        type = task[-1]
        # create a dict entry for the edge name
        edge_label_dict[(input, output)] = edge_label
        # createa a dict entry for the node name
        if type == "processing":
            node_label_dict[input] = input
            node_label_dict[output] = output
        else:
            node_label_dict[input] = input
            node_label_dict[output] = ""

    plt.figure(figsize=(x_log, y_log))
    G=nx.from_pandas_edgelist(df, 'input', 'output', create_using=nx.Graph())
    nx.draw(G, with_labels=False, node_color="skyblue", node_size=50, width=3, edge_color = "skyblue", edge_cmap=plt.cm.Blues, pos=nx.kamada_kawai_layout(G), font_size=8)
    nx.draw_networkx_edge_labels(G, pos=nx.kamada_kawai_layout(G), edge_labels=edge_label_dict, font_color='blue', font_size=6)
    nx.draw_networkx_labels(G, pos=nx.kamada_kawai_layout(G), labels=node_label_dict, font_color='black', font_size=8)
    plt.axis('off')
    plt.show(block=False)

    answer = sg.PopupYesNo('Save figure?', keep_on_top=True)
    if answer == "Yes":
        log_pdf = Path(str(path_to_outdirs) + "/log.pdf")
        plt.savefig(log_pdf)
        plt.close()
        closing_text = "Log network is found under:\n" + log_pdf.name
        sg.Popup(closing_text, title="Finished", keep_on_top=True)
    else:
        plt.close()



def replicate_consistency_filter(TaXon_table_xlsx, suffix_list, path_to_outdirs, consistency):

    import PySimpleGUI as sg
    import pandas as pd
    import numpy as np
    from pathlib import Path

    TaXon_table_xlsx = Path(TaXon_table_xlsx)
    TaXon_table_df = pd.read_excel(TaXon_table_xlsx)

    sample_names = TaXon_table_df.columns[10:].tolist()
    OTUs = TaXon_table_df["ID"].values.tolist()

    derep_sample_names_dict =  {}
    unique_sample_names_list = []
    replicates_dict = {}

    for sample in sample_names:
        sample_name = sample.split("_")[0:-1]
        unique_sample_names_list.append("_".join(sample_name))

    unique_sample_names_set = sorted(set(unique_sample_names_list))

    ############################################################################
    ## create the progress bar window
    layout = [[sg.Text('Progress bar')],
              [sg.ProgressBar(1000, orientation='h', size=(20, 20), key='progressbar')],
              [sg.Cancel()]]
    window_progress_bar = sg.Window('Progress bar', layout, keep_on_top=True)
    progress_bar = window_progress_bar['progressbar']
    progress_update = 0
    progress_increase = 1000 / len(unique_sample_names_set) + 1
    ############################################################################

    ## merge and replicate consistency version
    if consistency == True:
        no_replicates_list = []
        for sample in unique_sample_names_set:

            for i, suffix in enumerate(suffix_list):
                replicates_dict["rep_" + str(i)] = sample + "_" + str(suffix_list[i])

            replicate_names_list = list(replicates_dict.values())

            try:
                new_df = TaXon_table_df[replicate_names_list]
                header = new_df.columns.tolist()
                processed_reads = []

                for n_reads in new_df.values.tolist():
                    if 0 in n_reads:
                        if len(set(n_reads)) > 1:
                            n_reads = len(n_reads) * [0]
                    processed_reads.append(n_reads)

                df_out = pd.DataFrame(processed_reads)
                df_out.columns = header
                TaXon_table_df = TaXon_table_df.drop(replicate_names_list, axis=1)
                TaXon_table_df[sample] = df_out.sum(axis=1)

            except:
                no_replicates_list.append(sample)

            ############################################################################
            event, values = window_progress_bar.read(timeout=10)
            if event == 'Cancel'  or event is None:
                print('Cancel')
                window_progress_bar.Close()
                raise RuntimeError
            # update bar with loop value +1 so that bar eventually reaches the maximum
            progress_update += progress_increase
            progress_bar.UpdateBar(progress_update)
            ############################################################################

        window_progress_bar.Close()

        if len(no_replicates_list) == len(unique_sample_names_set):
            sg.PopupError("No replicates found. Please check your replicate suffixes.")
        else:
            dropped_OTUs_list = []
            # filter for 0 hit OTUs (can happen after consistency filtering)
            columns = TaXon_table_df.columns.tolist()
            TaXon_table_list = TaXon_table_df.values.tolist()
            TaXon_table_list_final = []
            for entry in TaXon_table_list:
                if sum(entry[10:]) != 0:
                    TaXon_table_list_final.append(entry)
                else:
                    print("Dropped:", entry[0], "(0 reads)")
                    dropped_OTUs_list.append(entry[0])

            taxon_tables_directory = Path(str(path_to_outdirs) + "/" + "TaXon_tables" + "/" + TaXon_table_xlsx.stem)
            output_xlsx = Path(str(taxon_tables_directory) + "_cons.xlsx")

            TaXon_table_df = pd.DataFrame(TaXon_table_list_final, columns=columns)
            TaXon_table_df.to_excel(output_xlsx, sheet_name='TaXon table', index=False)

            closing_text = "Taxon table is found under:\n" + '/'.join(str(output_xlsx).split("/")[-4:]) + "\n\n" + str(len(dropped_OTUs_list)) + " OTUs were removed."
            sg.Popup(closing_text, title="Finished", keep_on_top=True)

            from taxontabletools.create_log import ttt_log
            ttt_log("replicate consistency", "processing", TaXon_table_xlsx.name, output_xlsx.name, "consistency merged", path_to_outdirs)

    ## merge only version
    else:
        no_replicates_list = []
        for sample in unique_sample_names_set:

            for i, suffix in enumerate(suffix_list):
                replicates_dict["rep_" + str(i)] = sample + "_" + str(suffix_list[i])

            replicate_names_list = list(replicates_dict.values())

            try:
                new_df = TaXon_table_df[replicate_names_list]
                TaXon_table_df = TaXon_table_df.drop(replicate_names_list, axis=1)
                TaXon_table_df[sample] = new_df.sum(axis=1)
            except:
                no_replicates_list.append(sample)

            ############################################################################
            event, values = window_progress_bar.read(timeout=10)
            if event == 'Cancel'  or event is None:
                print('Cancel')
                window_progress_bar.Close()
                raise RuntimeError
            # update bar with loop value +1 so that bar eventually reaches the maximum
            progress_update += progress_increase
            progress_bar.UpdateBar(progress_update)
            ############################################################################

        window_progress_bar.Close()

        if len(no_replicates_list) == len(unique_sample_names_set):
            sg.PopupError("No replicates found. Please check your replicate suffixes.")

        else:
            taxon_tables_directory = Path(str(path_to_outdirs) + "/" + "TaXon_tables" + "/" + TaXon_table_xlsx.stem)
            output_xlsx = Path(str(taxon_tables_directory) + "_merged.xlsx")

            TaXon_table_df.to_excel(output_xlsx, sheet_name='TaXon table', index=False)

            closing_text = "Taxon table is found under:\n" + '/'.join(str(output_xlsx).split("/")[-4:])
            sg.Popup(closing_text, title="Finished", keep_on_top=True)

            from taxontabletools.create_log import ttt_log
            ttt_log("replicate merging", "processing", TaXon_table_xlsx.name, output_xlsx.name, "merged", path_to_outdirs)



def convert_to_presence_absence(TaXon_table_xlsx, path_to_outdirs):

    from pathlib import Path
    import PySimpleGUI as sg
    import  pandas as pd

    TaXon_table_xlsx =  Path(TaXon_table_xlsx)
    TaXon_table_df = pd.read_excel(TaXon_table_xlsx, header=0)
    # create presence/absence table
    presence_absence_list = []
    for col in TaXon_table_df.values.tolist():
        presence_absence_list.append(col[0:10] + [int(1) if reads != 0 else int(0) for reads in col[10:]])
    df_pa = pd.DataFrame(presence_absence_list)
    df_pa.columns = TaXon_table_df.columns.tolist()
    output_file = Path(str(path_to_outdirs) + "/" + "TaXon_tables" + "/" + TaXon_table_xlsx.stem + "_pa.xlsx")
    df_pa.to_excel(output_file, index=False, sheet_name = 'TaXon table')
    closing_text = "Presence absence tables is found in: " + str(path_to_outdirs) + "/TaXon_tables/"
    sg.Popup(closing_text, title="Finished", keep_on_top=True)

    from taxontabletools.create_log import ttt_log
    ttt_log("presence absence conversion", "processing", TaXon_table_xlsx.name, output_file.name, "nan", path_to_outdirs)



def per_taxon_analysis(TaXon_table_xlsx, height, width, taxonomic_level, path_to_outdirs, template, theme, font_size):

    import PySimpleGUI as sg
    import pandas as pd
    import numpy as np
    from pathlib import Path
    import plotly.graph_objects as go
    from plotly.subplots import make_subplots
    import itertools, webbrowser

    ## create a y axis title text
    taxon_title = taxonomic_level.lower()

    ## adjust taxonomic level if neccessary
    if taxonomic_level in ["ASVs", "ESVs", "OTUs", "zOTUs"]:
        taxon_title = taxonomic_level
        taxonomic_level = "ID"

    ## collect plot variables
    color1 = theme[0]
    color2 = theme[1]
    opacity_value = theme[2]
    height = int(height)
    width = int(width)

    ## load taxon table
    TaXon_table_xlsx = Path(TaXon_table_xlsx)
    TaXon_table_df = pd.read_excel(TaXon_table_xlsx)
    TaXon_table_df = TaXon_table_df.fillna("nan")

    ## collect the taxa to test on
    taxa = sorted(list(set([taxon for taxon in TaXon_table_df[taxonomic_level].values.tolist() if taxon != "nan"])))

    ## check if there are more than 8 taxa
    answer = "Yes"
    if len(taxa) > 8:
        answer = sg.PopupYesNo("There are more than 8 taxa detected. This can render the plot difficult to read. Continue anyway?")

    if answer == "Yes":
        ## collect the OTUs
        OTUs = TaXon_table_df["ID"].values.tolist()

        ## count the number of OTUs per taxon
        n_OTUs = [TaXon_table_df[taxonomic_level].values.tolist().count(taxon) for taxon in taxa]

        ## collect all OTUs on species level
        OTU_species = [OTU for OTU in TaXon_table_df[[taxonomic_level, "Species"]].values.tolist() if OTU[1] != "nan"]
        OTU_species.sort()
        OTU_species = list(k for k,_ in itertools.groupby(OTU_species))
        OTU_species = [OTU[0] for OTU in OTU_species]
        n_species = [OTU_species.count(taxon) for taxon in taxa]

        ## count reads for each taxon
        n_reads = []
        for taxon in taxa:
            n_reads.append(sum([sum(OTU[10:]) for OTU in TaXon_table_df[TaXon_table_df[taxonomic_level]==taxon].values.tolist()]))

        if (taxonomic_level == "Species" or taxonomic_level == "Genus"):
            x_values = ["<i>" + taxon + "</i>" for taxon in taxa]
        else:
            x_values = taxa

        ## calculate the read proportions
        reads_sum = sum(n_reads)
        n_reads = [round(reads / reads_sum * 100, 2) for reads in n_reads]

        ## create subplots
        fig = make_subplots(rows=1, cols=2, subplot_titles=("A)", "B)"))

        ## percentage of reads per taxonomic level
        hovertext = 'Taxon: %{x}, Reads: %{y}'
        fig.add_trace(go.Bar(hovertemplate=hovertext, name="",x=x_values, y=n_reads),row=1, col=1)
        fig.update_yaxes(title_text = "reads (%)", title_standoff=5, row=1, col=1)
        fig.update_traces(marker_color=color1, marker_line_color=color2, marker_line_width=1, opacity=opacity_value, showlegend=False, row=1, col=1)

        ## Number of OTUs
        hovertext = 'Taxon: %{x}, OTUs: %{y}'
        title_text = "# " + taxon_title
        fig.add_trace(go.Bar(hovertemplate=hovertext, name="",x=x_values, y=n_OTUs, text=n_OTUs, showlegend=False),row=1, col=2)
        fig.update_traces(marker_color=color1, marker_line_color=color2, marker_line_width=1, opacity=opacity_value, row=1, col=2)
        fig.update_yaxes(title_text=title_text, title_standoff=5, row=1, col=2, rangemode="tozero")

        ## Number of OTUs on species level
        hovertext = 'Taxon: %{x}, Species: %{text}'
        fig.add_trace(go.Scatter(textposition = "top center", hovertemplate=hovertext, text=n_species, name="Species",x=x_values, y=n_OTUs, showlegend=True, mode='text'),row=1, col=2)
        fig.update_traces(marker_color=color1, marker_line_color=color2, marker_line_width=1, opacity=opacity_value, row=1, col=2)

        ## fig.add_annotation( text='─ Species', align='left', showarrow=False, xref='paper', yref='paper', x=1.05, y=0.5, bordercolor='black', borderwidth=1)

        ## update the layout
        fig.update_layout(barmode='stack', height=int(height), width=int(width), template=template, showlegend=True, font_size=font_size, title_font_size=font_size)
        fig.update_xaxes(tickmode='linear')
        fig.update_xaxes(tickangle=-90)

        ## write ouput files
        output_pdf = Path(str(path_to_outdirs) + "/Per_taxon_statistics/" + TaXon_table_xlsx.stem + "_" + taxonomic_level + ".pdf")
        output_html = Path(str(path_to_outdirs) + "/Per_taxon_statistics/" + TaXon_table_xlsx.stem + "_" + taxonomic_level + ".html")
        fig.write_image(str(output_pdf))
        fig.write_html(str(output_html))

        ## ask to show file
        answer = sg.PopupYesNo('Show plot?', keep_on_top=True)
        if answer == "Yes":
            webbrowser.open('file://' + str(output_html))

        ## print closing text
        closing_text = "Plots are found under: " + str(path_to_outdirs) + "/Per_taxon_statistics/"
        sg.Popup(closing_text, title="Finished", keep_on_top=True)

        ## write to log
        from taxontabletools.create_log import ttt_log
        ttt_log("per taxon statistics", "analysis", TaXon_table_xlsx.name, output_pdf.name, "", path_to_outdirs)



# check the input format
def check_taXon_table_format(taXon_table):

    import PySimpleGUI as sg
    import pandas as pd
    import numpy as np
    from pathlib import Path

    try:
        taXon_table_df = pd.read_excel(Path(taXon_table), "TaXon table")
        taXon_table_df = taXon_table_df.replace(np.nan, 'nan', regex=True)
    except:
        sg.PopupError("Could not find the TaXon table sheet", keep_on_top=True)
        raise


    ###################################
    # A) header prompt
    taXon_table_df_header = taXon_table_df.columns.tolist()[0:9]
    header_prompt = ["ID", "Phylum", "Class", "Order", "Family", "Genus", "Species", "Similarity", "Status"]

    if taXon_table_df_header != header_prompt:
        sg.PopupError("Oops! Something is wrong with the header!", title="Error", keep_on_top=True)
        raise RuntimeError("Oops! Something is wrong with the header!")

    ###################################
    # B) OTUs prompt

    ## removed in 1.1.6

    ###################################
    # C) Species prompt
    # does not necessarily need to include both the Genus and epithet, but it is recommended

    species_list = taXon_table_df['Species'].values.tolist()
    epithet_only_list = []

    for species in species_list:
        if (species != 'nan' and len(species.split()) < 2):
            epithet_only_list.append(species)

    if epithet_only_list != []:
        WarningMessage = "Warning: There are " + str(len(epithet_only_list)) +  " species that do not fit the binomial nomenclature." + "\n" + "This error message will be ignored, but it's recommended to use binomial nomenclature!"
        sg.Popup(WarningMessage, title="Warning", keep_on_top=True)


    ###################################
    # D) Samples prompt

    samples_list = taXon_table_df.columns.tolist()[10:]

    for sample in samples_list:
        if " " in sample:
            ErrorMessage = "Please do not use spaces in the sample names:\n" + sample
            sg.PopupError(ErrorMessage, title="Error", keep_on_top=True)
            raise RuntimeError(ErrorMessage)

    ###################################
    # E) Reads prompt

    OTU_reads_list = taXon_table_df.values.tolist()

    ask = False
    answer = False
    write = False

    for OTU_reads in OTU_reads_list:
        read_numbers = OTU_reads[10:]
        if sum(read_numbers) == 0:
            ask = True
        for read_number in read_numbers:
            try:
                read_number = int(read_number)
            except:
                OTU = OTU_reads[0]
                ErrorMessage = "Please check your read numbers in " + OTU + " -> " + str(read_number)
                sg.PopupError(ErrorMessage, title="Error", keep_on_top=True)
                raise RuntimeError(ErrorMessage)

    ###################################
    # F) Taxonomy consistency

    taxonomy_list = taXon_table_df[["ID", "Phylum", "Class", "Order", "Family", "Genus", "Species"]].values.tolist()
    for entry in taxonomy_list:
        if "nan" in entry:
            taxonomy = entry[1:]
            OTU = entry[0]
            n_nan = taxonomy.count("nan")
            if 6 - taxonomy.index("nan") != n_nan:
                answer = sg.PopupOKCancel("Internally missing taxonomy found!\nReplace with placeholder?\nWarning: This will overwrite the dataframe!")
                break
    if answer == 'OK':
        new_df_list = []
        for row in taXon_table_df.values.tolist():
            entry = row[0:7]
            if "nan" in entry:
                taxonomy = entry[1:]
                OTU = entry[0]
                n_nan = taxonomy.count("nan")
                if 6 - taxonomy.index("nan") != n_nan:
                    for item in [i for i,x in enumerate(entry) if x == "nan"]:
                        row[item] = "FLAG"
            row = ['' if x=='nan' else x for x in row]
            new_df_list.append(row)
        column_names = taXon_table_df.columns.tolist()
        taXon_table_df_new = pd.DataFrame(new_df_list, columns=column_names)
        write = True
        sg.Popup("Replaced missing taxonomy with a >FLAG< placeholder.\nPlease adjust the taxonomy in Excel.")

    if write == True:
        taXon_table_df_new.to_excel(taXon_table, sheet_name="TaXon table", index=False)


    sg.Popup("Your file looks great and is ready to use!", title="Taxonomy table check", keep_on_top=True)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="taxontabletools", # Replace with your own username
    version="1.3.0",
    author="Till-Hendrik Macher",
    author_email="till-hendrik.macher@uni-due.de",
    description="TaxonTableTools - A comprehensive, platform-independent graphical user interface software to explore and visualise DNA metabarcoding data",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pypa/taxontabletools",
    packages=setuptools.find_packages(),
    license = 'MIT',
    install_requires = ['pySimpleGUI>=4.15.2',
                        'pandas>=0.25.3',
                        'numpy>=1.18.1',
                        'matplotlib-venn>=0.11.5',
                        'xlrd>=1.2.0',
                        'openpyxl>=3.0.3',
                        'xlsxwriter>=1.2.7',
                        'biopython>=1.77',
                        'scikit-bio>=0.5.6',
                        'requests_html>=0.10.0',
                        'scipy>=1.5.1',
                        'shapely>=1.7.1',
                        'plotly>=4.9.0',
                        'kaleido>=0.0.3',
                        'statsmodels>=0.12.1',
                        'lastversion>=1.3.3',
                        'networkx>=2.4'],
    include_package_data = True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



def convert_to_perlodes(TaXon_table_xlsx, operational_taxon_list, path_to_outdirs):

    import PySimpleGUI as sg
    import pandas as pd
    from pandas import DataFrame
    import numpy as np
    from pathlib import Path

    #get the taxonomy from the operational taxon list
    operational_taxon_list_df = pd.read_excel(Path(operational_taxon_list), header=2, sheet_name="Operationelle Taxaliste")
    taxonomy_list = operational_taxon_list_df["Taxonname\n(Perlodes-Datenbank)"].values.tolist()
    taxonomy_list = [x for x in taxonomy_list if str(x) != 'nan']

    # get the according IDs from the operational taxon list
    IDs_list = operational_taxon_list_df["ID_\nART"].values.tolist()
    IDs_list = [x for x in IDs_list if str(x) != 'nan']

    # create a dict to store both the ID and the taxonomy
    operational_taxon_list_dict = {}
    for ID, taxonomy in zip(IDs_list, taxonomy_list):
       operational_taxon_list_dict[taxonomy] = int(ID)

    # load the taxon table and create a list
    TaXon_table_xlsx = Path(TaXon_table_xlsx)
    TaXon_table_df = pd.read_excel(TaXon_table_xlsx)
    TaXon_table_taxonomy = TaXon_table_df.columns.tolist()[0:7]
    samples_list = TaXon_table_df.columns.tolist()[10:]

    # store hits and dropped OTUs
    hit_list, dropped_list, transversion_list = [], [], []

    # loop through the taxon table
    for taxonomy in TaXon_table_df[TaXon_table_taxonomy].drop_duplicates().values.tolist():

       # collect the OTU name, species, genus and family and convert to perlodes format
       OTU = taxonomy[0]
       Species = taxonomy[6]
       Species_group = str(taxonomy[6]) + "-Gr."
       Genus = str(taxonomy[5]) + " sp."
       Family = str(taxonomy[4]) + " Gen. sp."

       # test if the OTU has a hit at: Species level, Genus level or Family level
       if Species in operational_taxon_list_dict.keys():
           # add to hit list
           hit_list.append([OTU] + [str(operational_taxon_list_dict[Species])] + [Species])
           # add to perlodes log file
           transversion_list.append(taxonomy + [str(operational_taxon_list_dict[Species])] + [Species])

       elif Species_group in operational_taxon_list_dict.keys():
           hit_list.append([OTU] + [str(operational_taxon_list_dict[Species_group])] + [Species_group])
           transversion_list.append(taxonomy + [str(operational_taxon_list_dict[Species_group])] + [Species_group])

       elif Genus in operational_taxon_list_dict.keys():
           hit_list.append([OTU] + [str(operational_taxon_list_dict[Genus])] + [Genus])
           transversion_list.append(taxonomy + [str(operational_taxon_list_dict[Genus])] + [Genus])

       elif Family in operational_taxon_list_dict.keys():
           hit_list.append([OTU] + [str(operational_taxon_list_dict[Family])] + [Family])
           transversion_list.append(taxonomy + [str(operational_taxon_list_dict[Family])] + [Family])

       # otherwise store the hit with an "nan"
       else:
           hit_list.append([OTU] + ["nan"])
           dropped_list.append(OTU)
           transversion_list.append(taxonomy + ["", ""])

    # create an output list for perlodes
    # make read abundaces binary
    perlodes_input_list = []
    for hit, row in zip(hit_list, TaXon_table_df.values.tolist()):
       # skip OTUs that were not in the OPT
       if hit[1] != "nan":
           reads_list = []
           # loop through all the sample of the file
           for reads in row[10:]:
               # reads > 0 --> 1
               if reads > 0:
                   reads_list.append(1)
               # reads == 0 --> 0
               else:
                   reads_list.append(0)
           # now append the taxonomy and the presence/absence to the perlodes list
           perlodes_input_list.append([hit[1]] + [hit[2]] + reads_list)

    # print the number of dropped OTUs
    number_of_initial_OTUs = len(TaXon_table_df[TaXon_table_taxonomy].drop_duplicates().values.tolist())
    print("Warning: Dropped " + str(len(dropped_list)) + " of " + str(number_of_initial_OTUs) + " OTUs.\n")

    # write the perlodes output file
    perlodes_df = pd.DataFrame(perlodes_input_list)
    perlodes_df.columns = ["ID_ART", "TAXON_NAME"] + samples_list

    # Perlocdes sums up the counts of all df_duplicates! So these must be removed
    perlodes_df.index = perlodes_df["TAXON_NAME"]
    # create a set of all present taxa
    taxa_set =  set(perlodes_df["TAXON_NAME"].values.tolist())
    perlodes_filtered_list = []

    # loop through all target taxa
    for taxon in taxa_set:
       # collect the ID and Taxon name
       ID_taxon_name = perlodes_df.loc[taxon][["ID_ART", "TAXON_NAME"]].values.tolist()[0]
       # if there are duplicates:
       try:
           # calculate the sum for each sample
           sum_of_pa = list(perlodes_df.loc[taxon][samples_list].sum())
           # start a new list for each taxon
           pa_list = ID_taxon_name
           # replace the count by either 1 or 0
           for pa in sum_of_pa:
               if pa > 0:
                   pa_list.append(1)
               else:
                   pa_list.append(0)
           perlodes_filtered_list.append(pa_list)
       # if there are no duplicates
       except:
           # simply take the old presence/absence line
           perlodes_filtered_list.append(perlodes_df.loc[taxon].values.tolist())

    # write the filtered list to a dataframe
    perlodes_df = pd.DataFrame(perlodes_filtered_list)
    perlodes_df.columns = ["ID_ART", "TAXON_NAME"] + samples_list
    perlodes_directory = Path(str(path_to_outdirs) + "/" + "Perlodes" + "/" + TaXon_table_xlsx.stem)
    perlodes_xlsx = Path(str(perlodes_directory) + "_perlodes.xlsx")
    perlodes_df.to_excel(perlodes_xlsx, sheet_name='ImportList', index=False)

    # write the log file file to a different dataframe
    transversion_df = pd.DataFrame(transversion_list)
    transversion_df.columns = ["ID", "Phylum", "Class", "Order", "Family", "Genus", "Species", "ID_ART", "TAXON_NAME"]
    transversion_xlsx = Path(str(perlodes_directory) + "_perlodes_conversion_table.xlsx")
    transversion_df.to_excel(transversion_xlsx, index=False)

    closing_text = "Perlodes input file is found under:\n" + '/'.join(str(perlodes_xlsx).split("/")[-4:])
    sg.Popup(closing_text + "\n\nWarning: Please check the converted table for errors and compare the conversion results in the conversion table.\n\nThe conversion is still in beta and might contain errors!", title="Finished", keep_on_top=True)

    from taxontabletools.create_log import ttt_log
    ttt_log("perlodes conversion", "processing", TaXon_table_xlsx.name, perlodes_xlsx.name, "nan", path_to_outdirs)



def NMDS_analysis(TaXon_table_xlsx, meta_data_to_test, taxonomic_level, width, height, nmds_s, max_iter_val, n_init_val, path_to_outdirs, template, font_size, color_discrete_sequence, nmds_dissimilarity):

    import pandas as pd
    import numpy as np
    from skbio.diversity import beta_diversity
    from sklearn.manifold import MDS
    import plotly.graph_objects as go
    import plotly.express as px
    from pathlib import Path
    import PySimpleGUI as sg
    import os, webbrowser
    from itertools import combinations

    TaXon_table_xlsx =  Path(TaXon_table_xlsx)
    Meta_data_table_xlsx = Path(str(path_to_outdirs) + "/" + "Meta_data_table" + "/" + TaXon_table_xlsx.stem + "_metadata.xlsx")
    TaXon_table_df = pd.read_excel(TaXon_table_xlsx, header=0).fillna("unidentified")
    TaXon_table_samples = TaXon_table_df.columns.tolist()[10:]
    Meta_data_table_df = pd.read_excel(Meta_data_table_xlsx, header=0)
    Meta_data_table_samples = Meta_data_table_df['Samples'].tolist()

    ## create a y axis title text
    taxon_title = taxonomic_level.lower()

    ## adjust taxonomic level if neccessary
    if taxonomic_level in ["ASVs", "ESVs", "OTUs", "zOTUs"]:
        taxon_title = taxonomic_level
        taxonomic_level = "ID"

    ## create a subfolder for better sorting and overview
    dirName = Path(str(path_to_outdirs) + "/" + "NMDS_plots" + "/" + TaXon_table_xlsx.stem + "/")
    if not os.path.exists(dirName):
        os.mkdir(dirName)

    # check if the meta data differs
    if len(set(Meta_data_table_df[meta_data_to_test])) == len(Meta_data_table_df['Samples'].tolist()):
        sg.Popup("The meta data is unique for all samples. Please adjust the meta data table!", title=("Error"))
        raise RuntimeError

    if sorted(TaXon_table_samples) == sorted(Meta_data_table_samples):

        samples = Meta_data_table_samples

        ## extract the relevant data
        TaXon_table_df = TaXon_table_df[[taxonomic_level] + samples]

        ## define an aggregation function to combine multiple hit of one taxonimic level
        aggregation_functions = {}
        ## define samples functions
        for sample in samples:
            ## 'sum' will calculate the sum of p/a data
            aggregation_functions[sample] = 'sum'
        ## define taxon level function
        aggregation_functions[taxonomic_level] = 'first'

        ## create condensed dataframe
        df_new = TaXon_table_df.groupby(TaXon_table_df[taxonomic_level]).aggregate(aggregation_functions)
        if 'unidentified' in df_new.index:
            df_new = df_new.drop('unidentified')

        ## collect reads
        data = df_new[samples].transpose().values.tolist()
        ## calculate jaccard distances
        jaccard_dm = beta_diversity(nmds_dissimilarity, data, samples)

        ## NMDS function
        def nmds_function(matrix, dimensions):
            nmds = MDS(n_components=dimensions, metric=False, dissimilarity='precomputed', max_iter=int(max_iter_val), n_init=int(n_init_val))
            nmds_results = nmds.fit(jaccard_dm[:100])
            stress = round(nmds_results.stress_, 2)
            nmds_array = nmds_results.embedding_
            return({"stress":stress,"nmds_results":nmds_array})


        answer = sg.PopupOKCancel("The NMDS calculation may take a while. Continue?")

        if answer == "OK":
            ## test different dimensions
            nmds_results_dict = {}
            stress_dict = {}
            for i in range(1,11):
                nmds_results = nmds_function(jaccard_dm, i)
                nmds_results_dict[i] = nmds_results
                stress_dict[i] = nmds_results["stress"]

            ####################################################################################################
            win2_active = True

            layout2 = [[sg.Text("NMDS analysis options", size=(20,1))],
                    [sg.CB("Show stress plot", default=True, key="stress_plot")],
                    [sg.CB("Show NMDS 2D plot", default=True, key="2d_plot")],
                    [sg.CB("Show NMDS 3D plot", default=True, key="3d_plot")],
                    [sg.CB("Connect categories", default=True, key="draw_mesh")],
                    [sg.Text("")],
                    [sg.Button("Apply")]]

            win2 = sg.Window('NMDS analysis', layout2, keep_on_top=False)

            while True:
                event2, values2 = win2.Read()
                if event2 is None or event2 == 'Apply':
                    win2.close()
                    win2_active = False
                    break

            ####################################################################################################

            ## plot stress and dimensions
            fig = go.Figure()
            fig.add_trace(go.Scatter(x=list(stress_dict.keys()), y=list(stress_dict.values()), mode='markers+lines', name=sample, marker=dict(color="Blue", size=int(10))))
            fig.update_layout(showlegend=False, xaxis_title="Dimensions", yaxis_title="Stress")
            fig.update_layout(height=int(600), width=int(800), template=template, showlegend=False, font_size=font_size, title_font_size=font_size)

            ## define output files
            output_pdf = Path(str(dirName) + "/" + meta_data_to_test + "_" + taxon_title + "_stress.pdf")
            output_html = Path(str(dirName) + "/" + meta_data_to_test + "_" + taxon_title + "_stress.html")
            ## write output files
            fig.write_image(str(output_pdf))
            fig.write_html(str(output_html))
            ## ask to show file
            if values2['stress_plot'] == True:
                webbrowser.open('file://' + str(output_html))

            ####################################################################################################

            ## plot 2D
            stress = stress_dict[2]
            if values2["draw_mesh"] == True:
                ## create dataframe from NMDS results
                nmds_results_df = pd.DataFrame(nmds_results_dict[2]["nmds_results"], index=[samples])
                nmds_results_df.rename(columns={0:'NMDS1', 1:'NMDS2'}, inplace=True)
                nmds_results_df["Sample"] = samples
                nmds_results_df[meta_data_to_test] = Meta_data_table_df[meta_data_to_test].values.tolist()

                combinations_list =[]
                for metadata in nmds_results_df[meta_data_to_test]:
                    ## collect all entries for the respective metadata
                    arr = nmds_results_df.loc[nmds_results_df[meta_data_to_test] == metadata][['NMDS1', 'NMDS2', meta_data_to_test, "Sample"]].to_numpy()
                    ## create a df for all possible combinations using itertools combinations
                    for entry in list(combinations(arr, 2)):
                        combinations_list.append(list(entry[0]))
                        combinations_list.append(list(entry[1]))

                ## create a dataframe to draw the plot from
                df = pd.DataFrame(combinations_list)
                df.columns = ['NMDS1', 'NMDS2', meta_data_to_test, "Sample"]

                ## plot NMDS
                fig = go.Figure()
                fig = px.scatter(df, x="NMDS1", y="NMDS2", hover_data=['Sample'], color=meta_data_to_test, color_discrete_sequence=color_discrete_sequence)
                fig.update_traces(marker_size=int(nmds_s), mode="markers+lines", line=dict(width=0.5))
                fig.update_layout(title="Stress=" + str(stress), yaxis_title="NMDS1", xaxis_title="NMDS2")
                fig.update_layout(height=int(height), width=int(width), template=template, showlegend=True, font_size=font_size, title_font_size=font_size)

            else:

                ## create dataframe from NMDS results
                nmds_results_df = pd.DataFrame(nmds_results_dict[2]["nmds_results"], index=[samples])
                nmds_results_df.rename(columns={0:'X', 1:'Y'}, inplace=True)
                nmds_results_df[meta_data_to_test] = Meta_data_table_df[meta_data_to_test].values.tolist()
                nmds_results_df["Sample"] = samples
                ## plot NMDS
                fig = go.Figure()
                fig = px.scatter(nmds_results_df, x="X", y="Y", hover_data=['Sample'], color=meta_data_to_test, color_discrete_sequence=color_discrete_sequence)
                fig.update_traces(marker_size=int(nmds_s), mode="markers")
                fig.update_layout(title="Stress=" + str(stress), yaxis_title="NMDS1", xaxis_title="NMDS2")
                fig.update_layout(height=int(height), width=int(width), template=template, showlegend=True, font_size=font_size, title_font_size=font_size)

            ## define output files
            output_pdf = Path(str(dirName) + "/" + meta_data_to_test + "_" + taxon_title + "_2d.pdf")
            output_html = Path(str(dirName) + "/" + meta_data_to_test + "_" + taxon_title + "_2d.html")
            ## write output files
            fig.write_image(str(output_pdf))
            fig.write_html(str(output_html))
            ## ask to show file
            if values2['2d_plot'] == True:
                webbrowser.open('file://' + str(output_html))

            ####################################################################################################

            ## plot 3D
            stress = stress_dict[3]

            if values2["draw_mesh"] == True:
                ## create dataframe from NMDS results
                nmds_results_df = pd.DataFrame(nmds_results_dict[3]["nmds_results"], index=[samples])
                nmds_results_df["Sample"] = samples
                nmds_results_df[meta_data_to_test] = Meta_data_table_df[meta_data_to_test].values.tolist()
                nmds_results_df.rename(columns={0:'NMDS1', 1:'NMDS2', 2:'NMDS3'}, inplace=True)

                combinations_list =[]
                for metadata in nmds_results_df[meta_data_to_test]:
                    ## collect all entries for the respective metadata
                    arr = nmds_results_df.loc[nmds_results_df[meta_data_to_test] == metadata][['NMDS1', 'NMDS2', 'NMDS3', meta_data_to_test, "Sample"]].to_numpy()
                    ## create a df for all possible combinations using itertools combinations
                    for entry in list(combinations(arr, 2)):
                        combinations_list.append(list(entry[0]))
                        combinations_list.append(list(entry[1]))

                ## create a dataframe to draw the plot from
                df = pd.DataFrame(combinations_list)
                df.columns = ['NMDS1', 'NMDS2', 'NMDS3', meta_data_to_test, "Sample"]

                ## plot NMDS
                fig = go.Figure()
                ## draw the plot
                fig = px.scatter_3d(df, x="NMDS1", y="NMDS2", z="NMDS3", color=meta_data_to_test, text="Sample", title="textbox", color_discrete_sequence=color_discrete_sequence)
                fig.update_traces(marker_size=int(12), mode="markers+lines", line=dict(width=1))
                fig.update_layout(height=int(height), width=int(width), template=template, title="Stress=" + str(stress), showlegend=True, font_size=font_size, title_font_size=font_size)
                fig.update_layout(scene = dict(xaxis_title="NMDS1",yaxis_title="NMDS2",zaxis_title="NMDS3"))

            else:
                ## create dataframe from NMDS results
                nmds_results_df = pd.DataFrame(nmds_results_dict[3]["nmds_results"], index=[samples])
                nmds_results_df["Sample"] = samples
                nmds_results_df[meta_data_to_test] = Meta_data_table_df[meta_data_to_test].values.tolist()
                nmds_results_df.rename(columns={0:'NMDS1', 1:'NMDS2', 2:'NMDS3'}, inplace=True)
                ## plot NMDS
                fig = go.Figure()
                ## draw the plot
                fig = px.scatter_3d(nmds_results_df, x="NMDS1", y="NMDS2", z="NMDS3", color=meta_data_to_test, color_discrete_sequence=color_discrete_sequence)
                fig.update_traces(marker_size=int(12), mode="markers", line=dict(width=1))
                fig.update_layout(height=int(height), width=int(width), template=template, title="Stress=" + str(stress), showlegend=True, font_size=font_size, title_font_size=font_size)
                fig.update_layout(scene = dict(xaxis_title="NMDS1",yaxis_title="NMDS2",zaxis_title="NMDS3"))


            ## define output files
            output_pdf = Path(str(dirName) + "/" + meta_data_to_test + "_" + taxon_title + "_3d.pdf")
            output_html = Path(str(dirName) + "/" + meta_data_to_test + "_" + taxon_title + "_3d.html")
            ## write output files
            fig.write_image(str(output_pdf))
            fig.write_html(str(output_html))
            ## ask to show file
            if values2['3d_plot'] == True:
                webbrowser.open('file://' + str(output_html))

            ####################################################################################################

            ## print closing text
            closing_text = "NMDS plots are found in: " + str(path_to_outdirs) + "/NMDS_plots/"
            sg.Popup(closing_text, title="Finished", keep_on_top=True)
            ## write log file
            from taxontabletools.create_log import ttt_log
            ttt_log("nmds analysis", "analysis", TaXon_table_xlsx.name, output_pdf.name, meta_data_to_test, path_to_outdirs)



# remove PCR or Extraction Replicates
def combine_replicates(TaXon_table_xlsx, suffix_list, path_to_outdirs):

    import PySimpleGUI as sg
    import pandas as pd
    import numpy as np
    from pathlib import Path

    TaXon_table_file =  Path(TaXon_table_xlsx)

    # create output file
    output_file = Path(str(path_to_outdirs) + "/" + "TaXon_tables" + "/" + str(TaXon_table_file.stem) + "_derep.xlsx")
    TaXon_table_xlsx = pd.ExcelFile(TaXon_table_xlsx)
    df = pd.read_excel(TaXon_table_xlsx, 'TaXon table', header=0)

    sample_names = df.columns[10:]
    unique_sample_names_list, samples_to_process_list = [], []

    for sample in sample_names:
        sample_name = sample.split("_")[0:-1]
        unique_sample_names_list.append("_".join(sample_name))

    unique_sample_names_set = sorted(set(unique_sample_names_list))

    replicates_dict ={}

    for sample in unique_sample_names_set:

        for i, suffix in enumerate(suffix_list):
            replicates_dict["rep_" + str(i)] = sample + "_" + suffix_list[i]

        combined = sample + "_comb"

        replicate_names_list = list(replicates_dict.values())

        try:
            df[combined] = df[replicate_names_list].sum(axis=1)
            df = df.drop(replicate_names_list, axis=1)
        except:
            print("Warning! No replicates found for:    " + sample)

    df.to_excel(output_file, index=False, sheet_name = 'TaXon table')
    closing_text = "Taxon table is found under:\n" + '/'.join(str(output_file).split("/")[-4:])
    sg.Popup(closing_text, title="Finished", keep_on_top=True)

    from taxontabletools.create_log import ttt_log
    ttt_log("replicate merging", "processing", TaXon_table_file.name, output_file.name, "nan", path_to_outdirs)



def replicate_correlation_analysis(TaXon_table_xlsx, suffix_list, path_to_outdirs, width, height, repcorr_options, template, theme, font_size, clustering_unit):

    import PySimpleGUI as sg
    import pandas as pd
    import numpy as np
    from pathlib import Path
    import numpy as np
    import scipy.stats, webbrowser, os
    import matplotlib.pyplot as plt
    import plotly.express as px
    from plotly.subplots import make_subplots
    import plotly.graph_objects as go
    import statsmodels.api as sm

    color1 = theme[0]
    color2 = theme[1]
    opacity_value = theme[2]

    width = int(width)
    height = int(height)
    x_zero = repcorr_options[0]
    y_zero = repcorr_options[1]

    TaXon_table_xlsx = Path(TaXon_table_xlsx)
    TaXon_table_df = pd.read_excel(TaXon_table_xlsx)

    ## create an output folder
    replicate_analysis_name = Path(TaXon_table_xlsx).name.replace(".xlsx", "")
    dirName = Path(str(path_to_outdirs) + "/Replicate_analysis/" + replicate_analysis_name)
    if not os.path.exists(dirName):
        os.mkdir(dirName)

    sample_names = TaXon_table_df.columns[10:].tolist()
    OTUs = TaXon_table_df["ID"].values.tolist()

    derep_sample_names_dict =  {}
    unique_sample_names_list = []
    replicates_dict = {}

    for sample in sample_names:
        sample_name = sample.split("_")[0:-1]
        unique_sample_names_list.append("_".join(sample_name))

    unique_sample_names_set = sorted(set(unique_sample_names_list))

    if len(suffix_list) == 2:

        ##############################
        # number of reads

        x1 = []
        y1 = []

        for sample in unique_sample_names_set:
            rep1, rep2 = [], []
            try:
                rep1 = (sample + "_" + suffix_list[0])
                rep2 = (sample + "_" + suffix_list[1])
                if (rep1 != [] and rep2 != []):
                    reads_rep1 = sum(TaXon_table_df[str(rep1)].values.tolist())
                    reads_rep2 = sum(TaXon_table_df[str(rep2)].values.tolist())
                    y1.append(reads_rep1)
                    x1.append(reads_rep2)
            except:
                print("No replicates found for:", sample)

        ##############################
        # number of OTUs

        x2 = []
        y2 = []

        for sample in unique_sample_names_set:
            rep1, rep2 = [], []
            try:
                rep1 = (sample + "_" + suffix_list[0])
                rep2 = (sample + "_" + suffix_list[1])
                if (rep1 != [] and rep2 != []):
                    OTUs_rep1 = len([OTU for OTU in TaXon_table_df[str(rep1)].values.tolist() if OTU != 0])
                    OTUs_rep2 = len([OTU for OTU in TaXon_table_df[str(rep2)].values.tolist() if OTU != 0])
                    y2.append(OTUs_rep1)
                    x2.append(OTUs_rep2)
            except:
                print("No replicates found for:", sample)

        # spearman's rho
        # reads
        spearman = scipy.stats.spearmanr(x1, y1)
        spearman_p_reads = spearman[1]
        if spearman_p_reads <= 0.05:
            spearman_rho_reads = str(round(spearman[0], 3)) + "*"
        else:
            spearman_rho_reads = str(round(spearman[0], 3))
        # OTUs
        spearman = scipy.stats.spearmanr(x2, y2)
        spearman_p_OTUs = spearman[1]
        if spearman_p_OTUs <= 0.05:
            spearman_rho_OTUs = str(round(spearman[0], 3)) + "*"
        else:
            spearman_rho_OTUs = str(round(spearman[0], 3))

        # create subplots
        fig = make_subplots(rows=1, cols=2, subplot_titles=("A)", "B)"))
        # calculate line if best fit for reads and add the scatter plot
        df = pd.DataFrame({'X':x1, 'Y':y1})
        df['bestfit'] = sm.OLS(df['Y'],sm.add_constant(df['X'])).fit().fittedvalues
        fig.add_trace(go.Scatter(name="Reads",x=x1, y=y1, mode='markers', marker=dict(color=color1), showlegend=False),row=1, col=1)
        fig.add_trace(go.Scatter(name="rho(reads)=" + spearman_rho_reads, x=x1, y=df['bestfit'], mode='lines', marker=dict(color=color2)),row=1, col=1)
        fig.update_xaxes(title_text = "# reads (rep2)", row=1, col=1)
        fig.update_yaxes(title_text = "# reads (rep1)", row=1, col=1)
        # calculate line if best fit for OTUs and add the scatter plot
        df = pd.DataFrame({'X':x2, 'Y':y2})
        df['bestfit'] = sm.OLS(df['Y'],sm.add_constant(df['X'])).fit().fittedvalues
        fig.add_trace(go.Scatter(name=clustering_unit, x=x2, y=y2, mode='markers', marker=dict(color=color1), showlegend=False),row=1, col=2)
        title1 = "rho(" + clustering_unit + ")="+  spearman_rho_OTUs
        fig.add_trace(go.Scatter(name=title1, x=x2, y=df['bestfit'], mode='lines', marker=dict(color=color2)),row=1, col=2)
        title2 = "# " + clustering_unit + " (rep2)"
        fig.update_xaxes(title_text = title2, row=1, col=2)
        title3 = "# " + clustering_unit + " (rep3)"
        fig.update_yaxes(title_text = title3, row=1, col=2)
        # update the layouts
        fig.update_layout(height=int(height), width=int(width), template=template, font_size=font_size, title_font_size=font_size, showlegend=True)
        if y_zero == True:
            fig.update_yaxes(rangemode="tozero")
        if x_zero == True:
            fig.update_xaxes(rangemode="tozero")

        ## write files
        output_pdf = Path(str(dirName) + "/" + TaXon_table_xlsx.stem + "_repcorr_OTUs.pdf")
        output_html = Path(str(dirName) + "/" + TaXon_table_xlsx.stem + "_repcorr_OTUs.html")
        output_text = Path(str(dirName) + "/" + TaXon_table_xlsx.stem + "_repcorr_OTUs.txt")

        fig.write_image(str(output_pdf))
        fig.write_html(str(output_html))
        f = open(output_text, "w")
        f.write("Spearman correlation results\n")
        f.write("Reads\n" + "rho = " + str(spearman_rho_reads) + "\np = " + str(spearman_p_reads) + "\n")
        f.write(clustering_unit + "\n" + "rho = " + str(spearman_rho_OTUs) + "\np = " + str(spearman_p_OTUs) + "\n")
        f.close()

        ## ask to show file
        answer = sg.PopupYesNo('Show plot?', keep_on_top=True)
        if answer == "Yes":
            webbrowser.open('file://' + str(output_html))

        ## print closing text
        closing_text = "Plots are found under:\n" + "Projects/Replicate_analysis/"
        sg.Popup(closing_text, title="Finished", keep_on_top=True)

        ## write to log
        from taxontabletools.create_log import ttt_log
        ttt_log("replicate correlation analysis", "analysis", TaXon_table_xlsx.name, output_pdf.name, "", path_to_outdirs)

    else:
        sg.Popup("Correlation analyses can only be performed with two replicates.")
